from collections import namedtuple
from subprocess import Popen, PIPE


class Base:
    fields = None

    def tmux(self, cmd, *args, format=None, target=None):
        cmd = 'tmux', cmd
        if args:
            cmd+=args
        if format:
            cmd+=('-F', format)
        if target:
            cmd+=('-t', target)

        try:
            proc = Popen(cmd, stdout=PIPE, stderr=PIPE)
            stdout, stderr = proc.communicate()
            ret = proc.returncode
        except Exception as x:
            print(f'\n ! {" ".join(cmd)} \n ! {x}\n')
            raise RuntimeError

        err = stderr.decode().strip()
        if err:
            print(f'\n ! {" ".join(cmd)} [{ret}]\n ! {err}\n')
            raise RuntimeError

        lines = [line.strip() for line in stdout.decode().split('\n') if line.strip()]

        return lines

    def __getattr__(self, name):
        val = getattr(self.fields, name, None)
        if val is None:
            val = getattr(self.active, name)
        return val

    @classmethod
    def spec(cls):
        if not hasattr(cls, 'nt'):
            keys = [
                k.strip()
                for k in cls.Child.keys.split()
                if k.strip()
            ]
            cls.nt = namedtuple('fields', keys)
        return cls.nt._fields, cls.nt._make

    def format(self, keys):
        fl = ['#{%s}' % k for k in keys]
        sep = '\t'
        return sep.join(fl)

    @property
    def children(self):
        keys, make = self.spec()
        if isinstance(self.list_cmd, tuple):
            cmd, arg = self.list_cmd
            lines = self.tmux(
                cmd, arg,
                format = self.format(keys),
            )
        else:
            lines = self.tmux(
                self.list_cmd,
                format = self.format(keys),
            )
        for line in lines:
            vals = line.split('\t')
            fields = make(vals)
            yield self.Child(self, fields)

    @property
    def is_active(self):
        isactive = f'{self.__class__.__name__.lower()}_active'
        return getattr(self, isactive)=='1'

    @property
    def active(self):
        for child in self.children:
            if child.is_active:
                return child



from time import sleep

from teamux import tmux


def test_s_window_new():
    w = tmux.active.new('test_new')
    w.close()


def test_n_window_split():
    w = tmux.active.new('test_split')
    w.split()
    w.close()


def test_f_pane_resize():
    w = tmux.active.new('test_resize')
    w.split()
    w.right.width = 20
    sleep(2)
    w.left.width = 30
    sleep(2)
    w.left.width = 90
    w.close()


def test_f_pane_zoom():
    w = tmux.active.new('test_resize')
    w.split()
    w.right.send('# RIGHT')
    w.left.send('# LEFT')
    sleep(2)
    w.right.zoom()
    sleep(2)
    w.right.zoom()
    w.left.zoom()
    sleep(2)
    w.left.zoom()
    sleep(2)
    w.close()


def test_f_pane_print():
    w = tmux.active.new('test_print')
    w.split()
    for pane in w.panes:
        print(pane)
    w.close()

test_f_pane_print()



from .base import Base
from .session import Session


class Tmux(Base):
    Child = Session
    list_cmd = 'list-sessions'

    @property
    def sessions(self):
        yield from self.children

    def get(self, name):
        for s in self.sessions:
            if s.session_name==name:
                return s

    def new(self, name='xs'):
        self.tmux('new-session', '-s', name)
        return self.get(name)

    def __str__(self):
        return 'tmux'

    def tree(self):
        print('tmux')
        for s in self.sessions:
            print(' '*3, s)
            for w in s.windows:
                print(' '*6, w)
                for p in w.panes:
                    print(' '*9, p)


tmux = Tmux()



from time import sleep

from .base import Base
from .pane import Pane


class Window(Base):
    keys = '''
        session_id
        window_id
        window_name
        window_active
        window_width
        window_height
        window_index
    '''
    Child = Pane
    list_cmd = 'list-panes', '-a'


    def __init__(self, session, fields):
        self.session = session
        self.fields = fields

    @property
    def width(self):
        return int(self.window_width)

    @property
    def panes(self):
        for pane in self.children:
            if (pane.session_id==self.session_id
                and pane.window_id==self.window_id
            ):
                yield pane

    def select(self, target):
        self.tmux(f'select-pane', target=str(target))
        return self.active

    def split(self, hv='h'):
        self.tmux(f'split-window', f'-{hv}')
        sleep(1)

    def even(self):
        self.tmux('select-layout', 'even-horizontal')

    @property
    def right(self):
        return self.select('right')

    @property
    def left(self):
        return self.select('left')

    def swap(self):
        self.tmux(f'swap-pane', '-dU')

    def close(self):
        for pane in self.panes:
            pane.send('# Window closing ...')
            pane.close()

    def __str__(self):
        return f"{self.is_active and '*' or ' '} {self.window_name}"



from re import search
from time import sleep
from types import SimpleNamespace as ns

from .base import Base


class Pane(Base):
    keys = '''
        session_id
        window_id
        pane_id
        pane_active
        pane_index
        pane_width
        pane_height
        pane_title
        pane_pid
        pane_start_command
        pane_start_path
        pane_current_path
        pane_current_command
    '''

    def __init__(self, window, fields):
        self.window = window
        self.fields = fields
        self.prompt = r'.*\$$'
        self.context = None

    def send(self, keys, wait=True, dbg=False):
        self.tmux('send-keys', keys)
        self.tmux('send-keys', 'Enter')
        wait and self.wait(self.prompt, dbg=dbg)

    def send_keys(self, keys):
        self.tmux('send-keys', keys)

    def zoom(self):
        self.tmux('resize-pane', '-Z')

    @property
    def width(self):
        return int(self.pane_width)

    @width.setter
    def width(self, percent):
        w = self.window.width
        val = int(w*percent/100)
        self.tmux('resize-pane', f'-x{val}')

    @property
    def capture(self):
        return self.tmux('capture-pane', '-p')

    def search(self, pat, line, num):
        found = search(pat, line)
        if found:
            grp = found.groupdict()
            grp['line'] = line
            grp['num'] = num
            return ns(**grp)

    def parse(self, pat, num=-1):
        line = self.capture[num]
        return self.search(pat, line, num)

    def find(self, pat):
        num = 0
        for line in reversed(self.capture):
            num-=1
            found = self.search(pat, line, num)
            if found:
                return found

    def find_all(self, pat):
        for num, line in enumerate(self.capture):
            found = self.search(pat, line, num)
            if found:
                yield found

    def wait(self, pat, msg=None, back=1, dbg=None):
        go = True
        while go:
            sleep(.1)
            out = self.capture
            if not out: continue
            for line in out[-back:]:
                if search(pat, line):
                    if dbg:
                        print(f'{dbg} : {pat} found in {line}')
                    go = False
                else:
                    if dbg:
                        print(f'{dbg} : {pat} not in {line}')
        if msg:
            print(msg)

    def expect(self, back=1, **expected):
        go = True
        while go:
            sleep(.1)
            out = self.capture
            if not out: continue
            for case, pat in expected.items():
                for line in out[-back:]:
                    if search(pat, line):
                        return case

    def still(self, back=1, since=1):
        last = '__none__'
        while True:
            sleep(1)
            out = self.capture[-back]
            if out==last:
                count+=1
                if count==since: break
            else:
                count = 0
                last = out

    def run(self, cmd, exp=None, head=None, foot=None, back=1, dbg=None, raw=False):
        if head: print(head)
        self.send_keys(cmd) if raw else self.send(cmd, wait=False)
        if exp: self.wait(exp, foot, back, dbg)

    def script(self, source, exp=None, head=None, foot=None, back=1, dbg=None, wait=True, **kw):
        if head: print(head)

        source = source.format(**kw)
        for line in source.split('\n'):
            line = line.strip()
            if line and not line.startswith('#'):
                self.send(line, wait, dbg)

        if exp:
            self.wait(exp, foot, back, dbg)
        elif foot:
            print(foot)

    def close(self):
        self.send('# Pane closing ...', wait=False)
        sleep(1)
        self.send_keys('C-d')

    def __str__(self):
        lineage = f' > {self.context.lineage}' if self.context else ''
        return f"{self.is_active and '*' or ' '} {self.pane_title}{self.pane_id}{lineage}"




from .base import Base
from .window import Window


class Session(Base):
    keys = '''
        session_id
        session_name
        session_attached
    '''
    Child = Window
    list_cmd ='list-windows'

    def __init__(self, tmux, fields):
        self.Tmux = tmux
        self.fields = fields

    @property
    def windows(self):
        yield from self.children

    def get(self, name):
        for w in self.windows:
            if w.window_name==name:
                return w

    def new(self, name='xw'):
        self.tmux(f'new-window', '-n', name)
        return self.get(name)

    @property
    def is_active(self):
        return getattr(self, 'session_attached')=='1'

    def __str__(self):
        f = self.fields
        return f"{self.is_active and '*' or ' '} {f.session_name}"




from . import tmux


class Context:
    def __init__(self, manager, pane, parent, **environ):
        self.manager = manager
        self.pane = pane
        self.parent = parent
        self.environ = environ
        self.previous = None
        self.echo = True

    @property
    def lineage(self):
        return f'{self.parent.lineage}.{self.name}' if self.parent else self.name

    def do_enter(self):
        raise NotImplemented

    def run(self, cmd, **kw):
        head = kw.get(
            'head',
            f' -> {cmd}' if self.echo else None
        )
        prompt = kw.get('exp', self.prompt)
        self.pane.run(cmd, exp=prompt, head=head, **kw)

    def enter(self):
        current = self.manager.current
        print(f' + {self.name}')
        bak, self.echo = self.echo, False
        self.do_enter()
        self.pane.prompt = self.prompt
        self.echo = bak
        self.previous = current
        self.manager.current = self

    def exit_cmd(self):
        self.pane.send_keys('C-d')

    def wait_cmd(self, what):
        self.pane.wait(what)

    def exit(self):
        print(f' - {self.name}')
        self.exit_cmd()
        if self.previous:
            self.wait_cmd(self.previous.prompt)
            self.manager.current = self.previous
            self.pane.prompt = self.previous.prompt

    def __str__(self):
        return self.lineage


class Root(Context):
    name = None
    prompt = r'.*\$$'
    def do_enter(self): pass
    def exit_cmd(self): pass
    def exit(self): pass
    def __bool__(self):
        return False


class Manager:
    def __init__(self, pane, ctxs, **kw):
        prompt = kw.pop('prompt', None)
        if prompt:
            Root.prompt = prompt
        self.current = Root(
            self,
            pane,
            None,
        )

        cdict = {None : self.current}
        for ctx in ctxs:
            cdict[ctx.name] = ctx(
                self,
                pane,
                cdict.get(ctx.parent),
                **kw
            )
        self.contexts = cdict

    def enter(self, target):
        last = self.current

        target = self.contexts[target]
        targets = target.lineage.split('.')

        while self.current and self.current.name not in targets:
            self.current.exit()

        for name in targets:
            names = self.current.lineage.split('.') if self.current else []
            if name not in names:
                ctx = self.contexts[name]
                ctx.enter()

        if self.current != last:
            print(f' = {self.current}')

    def exit(self, target=None):
        last = self.current

        if not target:
            target = self.current.previous.name

        while self.current.name!=target:
            self.current.exit()
            if not self.current.parent:
                self.current.exit()
                break

        if self.current != last:
            print(f' = {self.current}')

    def __getattr__(self, attr):
        return getattr(self.current.pane, attr)


class MultiPanes:
    # mandatory seq of Context's subclass
    contexts = list()

    # optional arbitrary data transfered to contexts
    environ = dict()

    # space speparated panes's names
    panes = ''

    def __init__(self):
        assert self.contexts
        assert all(
            issubclass(ctx, Context)
            for ctx in self.contexts
        )
        self.__dict__.update(self.environ)
        self.names = {
            name : idx
            for idx, name in enumerate(self.panes.split())
        }
        assert len(names)>1

        session = tmux.active
        self.win = session.new()
        self.cur = self.win.left

        manager = Manager(
            self.cur,
            self.contexts,
            **self.environ
        )
        self.managers = [manager]
        self.current = manager

        for n in list(self.names)[1:]:
            self.add()
            self.win.even()

    def add(self):
        self.cur = self.win.right
        self.win.split()
        self.cur = self.win.right
        manager = Manager(
            self.cur,
            self.contexts,
            **self.environ
        )
        self.managers.append(manager)
        self.current = manager

    def select(self, name):
        idx = self.names[name]
        self.current = self.managers[idx]
        self.win.select(idx)

    @property
    def width(self):
        return self.current.current.pane.width

    @width.setter
    def width(self, percent):
        self.current.current.pane.width = percent

    def __getattr__(self, attr):
        try:
            return getattr(self.current, attr)
        except Exception as x:
            print(f'\n ! getattr {attr} ! {x}\n')
            raise

